/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * History:
 * 31-Oct-01 fhk; modified to support ldap check via bind on dn of user.
 * --------------------------------------------------------------------------*/
#include <jsm.h>
#include <lber.h>
#include <ldap.h>

#define TERROR_LDAPAUTH          (terror) {450,"Not Member"}

#define LDAPCHK_SUCCESS 1 /* TRUE */ 
#define LDAPCHK_FAIL    0 /* FALSE */

/* -------------------------------------------------------- */
/* Adjust these setting for your own LDAP server            */
/* these are loaded using the configurations in jabber.xml. */
/* -------------------------------------------------------- */
char HOSTNAME[50]; // the ldap server host name.
int PORT_NUMBER;   // the ldap port number.
char BASE_DN[64];  // the base dn for ldap searches.
char UIDATTR[50];  // the attribute that is used to map jabber login name to dn.

/* 
	Check ldap attribute.
		Connects to membership directory, finds t and checks the value of one attribute.
*/
short getLdapAttr(short *cause, char *filter, char *ctpasswd)
{
   short ret = LDAPCHK_FAIL;
   LDAP *ld;
   LDAPMessage *result, *e;
   char **vals;
   int rc;
   char *dn = NULL;
   char *attrs[] = {"userpassword", NULL};    // attributes to return from ldap search.

   *cause = 0;

   /* Get a handle to an LDAP connection. */
   if ((ld = ldap_init(HOSTNAME, PORT_NUMBER)) == NULL) 
   {  
      *cause = 1;
      log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 1 (ldap connect failed)");
      return LDAPCHK_FAIL;
   }

   /* Bind anonymously to the LDAP server. */
   if ((rc = ldap_simple_bind_s(ld, NULL, NULL)) != LDAP_SUCCESS) 
   {
      *cause = 2;
      log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 2 (ldap bind failed) on server %s, port %d",HOSTNAME,PORT_NUMBER);
      return LDAPCHK_FAIL;
   }

   log_debug(ZONE,"LDAP bind on server %s, port %d successful",HOSTNAME,PORT_NUMBER);
   
   /* Search for the entry. */
   log_debug(ZONE,"LDAP search ldap_search_s((ld, %s, LDAP_SCOPE_SUBTREE, %s, 'userpassword', LDAP_NO_LIMIT, &result)",BASE_DN,filter);
   if ((rc = ldap_search_s(ld, BASE_DN, LDAP_SCOPE_SUBTREE, filter, attrs, LDAP_NO_LIMIT, &result)) != LDAP_SUCCESS) 
   {
      *cause = 3;
      log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 3 (ldap search failed)");
      return LDAPCHK_FAIL;   
   } 

   /* Since we are doing a base search, there should be only one matching entry.  */
   e = ldap_first_entry(ld, result); 
   if (e != NULL) 
   {
     /* Get the attributes  */
     dn = ldap_get_dn(ld, e); // the DN
     log_debug(ZONE,"attempting login as dn=%s for %s using password %s",dn,filter,ctpasswd);
       
     if ((vals = ldap_get_values(ld, e, "userpassword")) != NULL) 
     {
        /* old way
        sprintf(ctpasswd,"%s", vals[0]);
        log_debug(ZONE,"val=%s len=%d  returning LDAPCHK_SUCCESS",ctpasswd, strlen(ctpasswd));
        ret = LDAPCHK_SUCCESS;
        */
        if ((rc = ldap_simple_bind_s(ld, dn, ctpasswd)) == LDAP_SUCCESS) {
         log_debug(ZONE,"returning LDAPCHK_SUCCESS for dn %s",dn);
         ret = LDAPCHK_SUCCESS;
        }else
        {
          *cause = 10;   
          log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 10 (invalid password)");
          ret = LDAPCHK_FAIL;
        }
     } 
     else 
     {
        *cause = 4;
        log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 4 (no such attribute, 'userpassword') for dn: %s",dn);
        ret = LDAPCHK_FAIL;
     } 

     ldap_value_free(vals);  
     ldap_memfree(dn);

   }
   else
   { 
     *cause = 11;
     log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 11 (user not found)");
     ret = LDAPCHK_FAIL;
           
   }

   ldap_msgfree(result);
   ldap_unbind(ld);

   return ret;
}

mreturn mod_auth_digest_ldap(mapi m, void *arg)
{
    spool s;
    char *sid;
    char *digest;
    char *mydigest;
    char *passxdb;
    char *filter = NULL;
    xmlnode xdb;
    short cause = 0;
    
    log_debug(ZONE,"start of auth control digest");

    if(jpacket_subtype(m->packet) == JPACKET__GET)
    { /* type=get means we flag that the server can do digest auth */
        xmlnode_insert_tag(m->packet->iq,"digest");
        log_debug(ZONE,"type is JPACKET__GET, inserted tag 'digest', returning M_PASS");
        return M_PASS;
    }

    if ((digest = xmlnode_get_tag_data(m->packet->iq,"digest")) == NULL){
      log_debug(ZONE,"'digest' not found in packet (m->packet->iq, returning M_PASS");
      return M_PASS;
    }

    log_debug(ZONE,"subtype=%d",jpacket_subtype(m->packet));
    
    /*  Check ldap records */
    filter = (char *) malloc(sizeof(char)*(strlen(m->user->id->user)+sizeof(char)*(strlen(UIDATTR))+4));
    sprintf(filter,"%s=%s",UIDATTR,m->user->id->user);

    xdb = xdb_get(m->si->xc, m->user->id, NS_AUTH);
    passxdb = xmlnode_get_data(xdb); // we hope this is the clear text password from the username.xml file.

    log_debug(ZONE,"user=%s password=%s(cleartext) filter=%s",m->user->id->user,passxdb,filter);
    
    if (getLdapAttr(&cause, filter, passxdb) == LDAPCHK_SUCCESS)
    {
           log_debug(ZONE,"LDAP OK cause=%d", cause);
           free(filter);
    }
    else
    {   
           jutil_error(m->packet->x, TERROR_LDAPAUTH);  
           log_debug(ZONE,"LDAP NOT FOUND cause=%d, returning M_HANDLED", cause);
           free(filter);
           return M_HANDLED;
    }

    xmlnode_free(xdb);

    /*
       hey, the rest of this code, at least the digest checking, is redundant, the call to getLdapAttr above performed
       exclusive check for authentication by binding on the clear text password (passxdb)
    */

    /* get the digest value from the stream */
    sid = xmlnode_get_attrib(xmlnode_get_tag(m->packet->iq,"digest"), "sid");

    /* Concat the stream id and password */
    /* SHA it up */
    log_debug(ZONE,"Got SID: %s", sid);
    s = spool_new(m->packet->p);
    spooler(s,sid,passxdb,s);
    mydigest = shahash(spool_print(s));

    log_debug(ZONE,"comparing %s %s",digest,mydigest);

    if(digest == NULL || sid == NULL || mydigest == NULL){
     log_debug(ZONE,"returning M_PASS");
     return M_PASS;
    }

    if(j_strcasecmp(digest, mydigest) != 0)
         {   jutil_error(m->packet->x, TERROR_LDAPAUTH);
             log_debug(ZONE,"compare failed");
         }
    else
        jutil_iqresult(m->packet->x);

    log_debug(ZONE,"returning M_HANDLED");
    return M_HANDLED;
}

void mod_auth_ldap_digest(jsmi si)
{
	xmlnode myx;
	char *pNode;

	log_debug(ZONE,"init");

	if ((myx = js_config(si, "ldapchk")) != NULL) 
	{
		log_debug(ZONE,"ldap settings:\n%s",xmlnode2str(myx));

		pNode = xmlnode_get_tag_data(myx,"host");
		strcpy(HOSTNAME, pNode);
		log_debug(ZONE,"hostname=%s",HOSTNAME);
		
		pNode = xmlnode_get_tag_data(myx,"port");
		PORT_NUMBER = atoi(pNode);
		log_debug(ZONE,"port=%d",PORT_NUMBER);
		
		pNode = xmlnode_get_tag_data(myx,"basedn");
		strcpy(BASE_DN, pNode);
		log_debug(ZONE,"basedn=%s",BASE_DN);

                pNode = xmlnode_get_tag_data(myx,"uidattr");
                strcpy(UIDATTR, pNode);
                log_debug(ZONE,"uidattr=%s",UIDATTR);
		
	}

	log_debug(ZONE,"calling js_mapi_register, e_AUTH, function: mod_auth_digest_ldap");
        js_mapi_register(si,e_AUTH, mod_auth_digest_ldap, NULL);
}
